from flask import request

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_token import AuthenticationToken


class PreAuthenticatedToken(AuthenticationToken):
    def __init__(self, principal, credentials=None, authorities=None):
        super().__init__(authorities=authorities)
        self._principal = principal
        self._credentials = credentials
        if authorities is not None:
            self.authenticate(True)

    @property
    def principal(self):
        return self._principal

    @property
    def credentials(self):
        return self._credentials


class BearerTokenExtractor:
    def extract(self) -> Authentication:
        token = self._extract_token_from_header()
        if token is None:
            token = self._extract_token_from_query()
        if token is not None:
            return PreAuthenticatedToken(token)

    @staticmethod
    def _extract_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.lower().startswith('bearer'):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _extract_token_from_query():
        token = request.args.get('access_token')
        if token is not None:
            return token



from abc import ABCMeta, abstractmethod

from guniflask.security_config.configured_security_builder import ConfiguredSecurityBuilder
from guniflask.web.request_filter import RequestFilter


class HttpSecurityBuilder(ConfiguredSecurityBuilder, metaclass=ABCMeta):

    @abstractmethod
    def add_request_filter(self, request_filter: RequestFilter):
        pass  # pragma: no cover



import datetime as dt

import sqlalchemy

from guniflask.utils.datatime import convert_to_datetime, local_tz_info
from guniflask.utils.rule import make_ignore_rule_for_field, make_include_rule_for_field, make_only_rule_for_field


class DictRecursionError(Exception):
    pass


def model_to_dict(model, ignore=None, only=None, include=None):
    return _model_to_dict(model, ignore=ignore, only=only, include=include)


def _model_to_dict(model, ignore=None, only=None, include=None, __prefix='', __exists=None) -> dict:
    if __exists is None:
        __exists = set()
    if model in __exists:
        raise DictRecursionError
    __exists.add(model)

    mapper = sqlalchemy.inspect(model).mapper
    col_attrs = mapper.column_attrs
    relationships = mapper.relationships

    d = {}
    ignore_set = make_ignore_rule_for_field(ignore)
    only_set = make_only_rule_for_field(only)
    include_set = make_include_rule_for_field(include)
    keys = list(col_attrs.keys())
    for k in relationships.keys():
        if _new_prefix(__prefix, k) in include_set:
            keys.append(k)
    for key in keys:
        if _in_set(__prefix, key, ignore_set):
            continue
        if only and not _in_set(__prefix, key, only_set):
            continue

        if key in col_attrs:
            v = getattr(model, key)
            if isinstance(v, dt.datetime) and v.tzinfo is None:
                v = v.replace(tzinfo=local_tz_info)
            d[key] = v
        elif key in relationships:
            if not hasattr(model, key):
                continue
            v = getattr(model, key)
            try:
                if v:
                    if isinstance(v, (list, set)):
                        t = [
                            _model_to_dict(
                                obj,
                                ignore=ignore_set,
                                only=only_set,
                                include=include_set,
                                __prefix=_new_prefix(__prefix, key),
                                __exists=__exists,
                            )
                            for obj in v
                        ]
                        v = t
                    else:
                        v = _model_to_dict(
                            v,
                            ignore=ignore_set,
                            only=only_set,
                            include=include_set,
                            __prefix=_new_prefix(__prefix, key),
                            __exists=__exists,
                        )
                d[key] = v
            except DictRecursionError:
                pass
    __exists.remove(model)
    return d


def result_to_dict(result, ignore=None, only=None) -> dict:
    res = {}
    ignore_set = make_ignore_rule_for_field(ignore)
    only_set = make_only_rule_for_field(only)
    for key in result.keys():
        if key not in ignore_set:
            if only and key not in only_set:
                continue
            v = getattr(result, key)
            if isinstance(v, dt.datetime) and v.tzinfo is None:
                v = v.replace(tzinfo=local_tz_info)
            res[key] = v
    return res


def dict_to_model(dict_obj: dict, model_cls, ignore=None, only=None):
    return _dict_to_model(dict_obj, model_cls, ignore=ignore, only=only)


def _dict_to_model(dict_obj: dict, model_cls, ignore=None, only=None, __prefix=''):
    mapper = sqlalchemy.inspect(model_cls).mapper
    col_attrs = mapper.column_attrs
    relationships = mapper.relationships
    ignore_set = make_ignore_rule_for_field(ignore)
    only_set = make_only_rule_for_field(only)
    kwargs = {}
    for key in dict_obj:
        if key in col_attrs or key in relationships:
            if _in_set(__prefix, key, ignore_set):
                continue
            if only and not _in_set(__prefix, key, only_set):
                continue
            v = dict_obj[key]
            if key in relationships:
                if isinstance(v, dict):
                    obj_cls = relationships[key].mapper.class_
                    v = _dict_to_model(
                        v,
                        obj_cls,
                        ignore=ignore_set,
                        only=only_set,
                        __prefix=_new_prefix(__prefix, key),
                    )
                elif isinstance(v, (list, set)):
                    obj_cls = relationships[key].mapper.class_
                    t = [
                        _dict_to_model(
                            obj,
                            obj_cls,
                            ignore=ignore_set,
                            only=only_set,
                            __prefix=_new_prefix(__prefix, key),
                        )
                        for obj in v
                    ]
                    v = t
            elif key in col_attrs:
                if isinstance(getattr(model_cls, key).type, sqlalchemy.DateTime):
                    v = convert_to_datetime(v)
            kwargs[key] = v
    model = model_cls(**kwargs)
    return model


def update_model_by_dict(model, dict_obj: dict, ignore=None, only=None):
    _update_model_by_dict(model, dict_obj, ignore=ignore, only=only)


def _update_model_by_dict(model, dict_obj: dict, ignore=None, only=None, __prefix=''):
    mapper = sqlalchemy.inspect(model).mapper
    col_attrs = mapper.column_attrs
    relationships = mapper.relationships
    ignore_set = make_ignore_rule_for_field(ignore)
    only_set = make_only_rule_for_field(only)
    for key in dict_obj:
        if key in col_attrs or key in relationships:
            if _in_set(__prefix, key, ignore_set):
                continue
            if only and not _in_set(__prefix, key, only_set):
                continue
            v = dict_obj[key]
            if key in relationships:
                if isinstance(v, dict):
                    obj = getattr(model, key)
                    if obj:
                        _update_model_by_dict(
                            obj,
                            v,
                            ignore=ignore_set,
                            only=only_set,
                            __prefix=_new_prefix(__prefix, key),
                        )
                        v = obj
                    else:
                        obj_cls = relationships[key].mapper.class_
                        v = _dict_to_model(
                            v,
                            obj_cls,
                            ignore=ignore_set,
                            only=only_set,
                            __prefix=_new_prefix(__prefix, key),
                        )
                elif isinstance(v, (list, set)):
                    # FIXME
                    raise RuntimeError(f'do not support to update a list: "{_new_prefix(__prefix, key)}"')
            elif key in col_attrs:
                if isinstance(col_attrs[key].class_attribute.type, sqlalchemy.DateTime):
                    v = convert_to_datetime(v)
            setattr(model, key, v)


def _in_set(prefix: str, key: str, s: set):
    return _new_prefix(prefix, key) in s


def _new_prefix(prefix: str, key: str):
    if not prefix:
        return key
    return prefix + '.' + key



from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.beans.factory import ConfigurableBeanFactory


class BeanFactoryPostProcessor:
    def post_process_bean_factory(self, bean_factory: ConfigurableBeanFactory):
        pass


class BeanDefinitionRegistryPostProcessor(BeanFactoryPostProcessor):
    def post_process_bean_definition_registry(self, registry: BeanDefinitionRegistry):
        pass



from .app_settings import Settings
from .app_settings import settings
from .env import app_name_from_env
from .env import load_app_env
from .env import set_app_default_env
from .load_utils import load_app_settings
from .load_utils import load_profile_config



from abc import ABCMeta

from guniflask.security.authentication import Authentication
from guniflask.security.user_details import UserDetails


class AuthenticationToken(Authentication, metaclass=ABCMeta):
    @property
    def name(self):
        if isinstance(self.principal, UserDetails):
            return self.principal.username
        if self.principal is None:
            return ''
        return str(self.principal)


class UserAuthentication(AuthenticationToken):
    def __init__(self, principal, credentials=None, authorities=None):
        super().__init__(authorities=authorities)
        self._principal = principal
        self._credentials = credentials
        if authorities is not None:
            self.authenticate(True)

    @property
    def principal(self):
        return self._principal

    @property
    def credentials(self):
        return self._credentials



import logging
from abc import ABCMeta, abstractmethod

from guniflask.beans.factory import BeanFactory
from guniflask.beans.factory_post_processor import BeanFactoryPostProcessor
from guniflask.context.event_listener import ApplicationEventListener

log = logging.getLogger(__name__)


class BeanContext(BeanFactory, metaclass=ABCMeta):

    @abstractmethod
    def add_application_listener(self, listener: ApplicationEventListener):
        pass  # pragma: no cover

    @abstractmethod
    def add_bean_factory_post_processor(self, post_processor: BeanFactoryPostProcessor):
        pass  # pragma: no cover

    @abstractmethod
    def refresh(self):
        pass  # pragma: no cover

    @abstractmethod
    def close(self):
        pass  # pragma: no cover


class BeanContextAware(metaclass=ABCMeta):
    @abstractmethod
    def set_bean_context(self, bean_context: BeanContext):
        pass  # pragma: no cover



from guniflask.security import has_role
from guniflask.web import blueprint, post_route, get_route

from ..services.account_service import AccountService


@blueprint
class AccountController:
    def __init__(self, accounts: AccountService):
        self.accounts = accounts

    @post_route('/login')
    def login(self, username: str, password: str):
        return self.accounts.login(username, password)

    @has_role('admin')
    @get_route('/accounts/<username>')
    def get_account_info(self, username: str):
        return self.accounts.get(username)



from abc import ABCMeta, abstractmethod

from guniflask.context.annotation import configuration, bean
from guniflask.scheduling.config_constants import *
from guniflask.scheduling.scheduling_post_processor import ScheduledPostProcessor
from guniflask.scheduling.task_scheduler import TaskScheduler, DefaultTaskScheduler


class SchedulingConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def get_task_scheduler(self) -> TaskScheduler:
        pass  # pragma: no cover


@configuration
class SchedulingConfiguration:
    def __init__(self, scheduled_configurer: SchedulingConfigurer = None):
        if scheduled_configurer is None:
            self._task_scheduler = DefaultTaskScheduler()
        else:
            self._task_scheduler = scheduled_configurer.get_task_scheduler()

    @bean(SCHEDULED_ANNOTATION_PROCESSOR)
    def scheduled_annotation_processor(self) -> ScheduledPostProcessor:
        return ScheduledPostProcessor()

    @bean
    def task_scheduler(self) -> TaskScheduler:
        return self._task_scheduler



from .base_model import BaseModelMixin
from .model_utils import dict_to_model
from .model_utils import model_to_dict
from .model_utils import result_to_dict
from .model_utils import update_model_by_dict
from .sqlalchemy_wrapper import wrap_sqlalchemy_model



from .authentication_manager_builder import AuthenticationManagerBuilder
from .http_security import HttpSecurity
from .http_security_builder import HttpSecurityBuilder
from .security_configurer import SecurityConfigurer
from .web_security import WebSecurity
from .web_security_configurer import WebSecurityConfigurer



APPLICATION_EVENT_PUBLISHER = '__application_event_publisher'
CONFIGURATION_ANNOTATION_PROCESSOR = '__configuration_annotation_processor'
CONFIGURATION_BEAN_NAME_GENERATOR = '__configuration_bean_name_generator'
AUTOWIRED_ANNOTATION_PROCESSOR = '__autowired_annotation_processor'



import json
import os
from os.path import isfile, join

import yaml


def _load_config(fname, **kwargs) -> dict:
    if fname is None or not isfile(fname):
        raise FileNotFoundError(f"Cannot find configuration file '{fname}'")

    s = {}
    if fname.endswith('.py'):
        s = _load_py_config(fname, **kwargs)
    if fname.endswith('.yaml') or fname.endswith('.yml'):
        s = _load_yaml_config(fname, **kwargs)
    if fname.endswith('.json'):
        s = _load_json_config(fname, **kwargs)

    _s = {}
    for name in s:
        if not name.startswith('_'):
            _s[name] = s[name]
    return _s


def _load_py_config(fname: str, **kwargs) -> dict:
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    cfg.update(kwargs)
    exec(code, cfg, cfg)
    return cfg


def _load_yaml_config(fname: str, **kwargs) -> dict:
    s = dict(**kwargs)
    with open(fname, 'rb') as f:
        s.update(yaml.safe_load(f))
    return s


def _load_json_config(fname: str, **kwargs) -> dict:
    s = dict(**kwargs)
    with open(fname, 'r') as f:
        s.update(json.load(f))
    return s


def load_profile_config(conf_dir: str, name: str, **kwargs) -> dict:
    _config_ext = ['.py', '.yaml', '.yml', '.json']

    pc = {}
    base_files = [join(conf_dir, name + ext) for ext in _config_ext]
    for file in base_files:
        try:
            pc = _load_config(file, **kwargs)
        except FileNotFoundError:
            pass
        else:
            break

    profiles = kwargs.get('active_profiles')
    if profiles:
        profiles = profiles.split(',')
        for profile in reversed(profiles):
            if profile:
                files = [join(conf_dir, name + '_' + profile + ext) for ext in _config_ext]
                for file in files:
                    try:
                        c = _load_config(file, **kwargs)
                    except FileNotFoundError:
                        pass
                    else:
                        _merge_config(pc, c)
                        break

    return pc


def _merge_config(old: dict, new: dict):
    for k, v in new.items():
        if k not in old:
            old[k] = v
        else:
            if isinstance(v, dict) and isinstance(old[k], dict):
                _merge_config(old[k], v)
            else:
                old[k] = v


def load_app_settings(app_name) -> dict:
    c = {}
    conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
    kwargs = get_settings_from_env()
    kwargs['app_name'] = app_name
    if conf_dir:
        c = load_profile_config(conf_dir, app_name, **kwargs)
    return c


def get_settings_from_env() -> dict:
    kwargs = {
        'home': os.environ.get('GUNIFLASK_HOME'),
    }
    if os.environ.get('GUNIFLASK_DEBUG'):
        kwargs['debug'] = True
    else:
        kwargs['debug'] = False
    kwargs['host'] = os.environ.get('GUNIFLASK_HOST')
    port = os.environ.get('GUNIFLASK_PORT')
    if port:
        port = int(port)
    kwargs['port'] = port
    active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
    kwargs['active_profiles'] = active_profiles
    return kwargs



from guniflask.orm.model_utils import model_to_dict, dict_to_model, update_model_by_dict


def wrap_sqlalchemy_model(model_cls):
    if not hasattr(model_cls, 'to_dict'):
        model_cls.to_dict = lambda self, **kwargs: model_to_dict(self, **kwargs)
    if not hasattr(model_cls, 'from_dict'):
        model_cls.from_dict = classmethod(lambda cls, dict_obj, **kwargs: dict_to_model(dict_obj, cls, **kwargs))
    if not hasattr(model_cls, 'update_by_dict'):
        model_cls.update_by_dict = lambda self, dict_obj, **kwargs: update_model_by_dict(self, dict_obj, **kwargs)
    return model_cls



import base64

from flask import request

from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.context import SecurityContext
from guniflask.security.errors import BadCredentialsError
from guniflask.security.web_authentication_details import WebAuthenticationDetails
from guniflask.web.request_filter import RequestFilter


class BasicAuthenticationFilter(RequestFilter):
    def __init__(self, authentication_manger: AuthenticationManager):
        self.authentication_manager = authentication_manger

    def before_request(self):
        header = request.headers.get('Authorization')
        if header is None or not header.startswith('Basic '):
            return
        username, password = self._extract_from_header(header)
        if self._require_authentication(username):
            auth_request = UserAuthentication(username, credentials=password)
            auth_request.details = WebAuthenticationDetails()
            auth_result = self.authentication_manager.authenticate(auth_request)
            SecurityContext.set_authentication(auth_result)

    def _extract_from_header(self, header: str):
        header = header[6:].encode('utf-8')
        decoded = base64.b64decode(header)
        token = decoded.decode('utf-8')
        s = token.split(':', maxsplit=1)
        if len(s) < 2:
            raise BadCredentialsError('Invalid basic authentication token')
        return s[0], s[1]

    def _require_authentication(self, username: str) -> bool:
        existing_auth = SecurityContext.get_authentication()
        if existing_auth is None or not existing_auth.is_authenticated:
            return True
        if isinstance(existing_auth, UserAuthentication) and existing_auth.name != username:
            return True
        return False



from abc import ABCMeta, abstractmethod

from apscheduler.schedulers.background import BackgroundScheduler

from guniflask.beans.lifecycle import DisposableBean, SmartInitializingSingleton


class AsyncExecutor(metaclass=ABCMeta):

    @abstractmethod
    def submit(self, task):
        pass  # pragma: no cover


class DefaultAsyncExecutor(AsyncExecutor, SmartInitializingSingleton, DisposableBean):
    def __init__(self, scheduler=None):
        if scheduler is None:
            try:
                from apscheduler.schedulers.gevent import GeventScheduler
                self._scheduler = GeventScheduler()
            except ImportError:
                self._scheduler = BackgroundScheduler()
        else:
            self._scheduler = scheduler

    @property
    def scheduler(self):
        return self._scheduler

    def submit(self, task):
        self._scheduler.add_job(task)

    def start(self):
        self._scheduler.start(paused=False)

    def shutdown(self, wait=True):
        self._scheduler.shutdown(wait=wait)

    def after_singletons_instantiated(self):
        self.start()

    def destroy(self):
        self.shutdown(wait=False)
